from typing import Any

from aiohttp import ClientResponseError


class MockResponse:
    def __init__(self, json_data: dict[str, Any], status_code: int, text: str = "") -> None:
        self._json = json_data
        self._text = text

        self.status = status_code

    async def json(self) -> dict[str, Any]:
        return self._json

    def raise_for_status(self) -> None:
        if self.status >= 400:
            raise ClientResponseError(
                None,  # type: ignore
                None,  # type: ignore
                status=self.status,
                message="Bad Request",
                headers=None,
            )



class HuumError(Exception):
    pass


class SafetyException(HuumError):
    pass


class BadRequest(HuumError):
    pass


class NotAuthenticated(HuumError):
    pass


class Forbidden(HuumError):
    pass


class RequestError(HuumError):
    pass



from typing import Any
from unittest import TestCase
from unittest.mock import patch

import pytest

from huum.exceptions import BadRequest, Forbidden, NotAuthenticated, RequestError
from huum.huum import Huum
from huum.schemas import HuumStatusResponse
from tests.utils import MockResponse


@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_status_idle(mock_request: Any) -> None:
    idle_status_response = {
        "maxHeatingTime": "3",
        "statusCode": 232,
        "door": True,
        "paymentEndDate": None,
        "temperature": "21",
        "saunaName": "test",
    }
    expected_result = HuumStatusResponse.from_dict(idle_status_response)
    mock_request.return_value = MockResponse(idle_status_response, 200)

    huum = Huum("test", "test")
    await huum.open_session()
    response = await huum.status()

    TestCase().assertDictEqual(response.to_dict(), expected_result.to_dict())


@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_status_heating(mock_request: Any) -> None:
    heating_status_response = {
        "maxHeatingTime": "3",
        "statusCode": 231,
        "door": True,
        "paymentEndDate": None,
        "temperature": "21",
        "targetTemperature": "75",
        "startDate": 1631623054,
        "endDate": 1631633854,
        "duration": 179,
        "saunaName": "test",
    }
    expected_result = HuumStatusResponse.from_dict(heating_status_response)
    mock_request.return_value = MockResponse(heating_status_response, 200)

    huum = Huum("test", "test")
    await huum.open_session()
    response = await huum.status()

    TestCase().assertDictEqual(response.to_dict(), expected_result.to_dict())


@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_heating_stop(mock_request: Any) -> None:
    heating_stop_response = {
        "maxHeatingTime": "3",
        "statusCode": 232,
        "door": True,
        "paymentEndDate": None,
        "temperature": "22",
        "targetTemperature": "75",
        "startDate": 1631685790,
        "endDate": 1631685790,
        "duration": 0,
        "saunaName": "test",
    }
    expected_result = HuumStatusResponse.from_dict(heating_stop_response)
    mock_request.return_value = MockResponse(heating_stop_response, 200)

    huum = Huum("test", "test")
    await huum.open_session()
    response = await huum.turn_off()

    TestCase().assertDictEqual(response.to_dict(), expected_result.to_dict())


@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_heating_start(mock_request: Any) -> None:
    heating_start_response = {
        "maxHeatingTime": "3",
        "statusCode": 231,
        "door": True,
        "paymentEndDate": None,
        "temperature": "22",
        "targetTemperature": "75",
        "startDate": 1631685780,
        "endDate": 1631696580,
        "duration": 180,
        "saunaName": "test",
    }
    expected_result = HuumStatusResponse.from_dict(heating_start_response)
    mock_request.return_value = MockResponse(heating_start_response, 200)

    huum = Huum("test", "test")
    await huum.open_session()
    response = await huum.turn_on(75)

    TestCase().assertDictEqual(response.to_dict(), expected_result.to_dict())


@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_toggle_light(mock_request: Any) -> None:
    toggle_light_response = {
        "maxHeatingTime": "3",
        "statusCode": 231,
        "door": True,
        "paymentEndDate": None,
        "temperature": "22",
        "targetTemperature": "75",
        "startDate": 1631685780,
        "endDate": 1631696580,
        "duration": 180,
        "saunaName": "test",
        "light": 1,
    }
    expected_result = HuumStatusResponse.from_dict(toggle_light_response)
    mock_request.return_value = MockResponse(toggle_light_response, 200)

    huum = Huum("test", "test")
    await huum.open_session()
    response = await huum.toggle_light()

    TestCase().assertDictEqual(response.to_dict(), expected_result.to_dict())


@pytest.mark.parametrize(
    (
        "status_code",
        "exception",
    ),
    [
        (
            400,
            BadRequest,
        ),
        (
            401,
            NotAuthenticated,
        ),
        (
            403,
            Forbidden,
        ),
        (
            500,
            RequestError,
        ),
    ],
)
@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_error_codes(mock_request: Any, status_code: int, exception: type[Exception]) -> None:
    mock_request.return_value = MockResponse({}, status_code)

    huum = Huum("test", "test")
    await huum.open_session()
    with pytest.raises(exception):
        await huum.status()



from huum.const import STATUS_CODE_TEXTS, SaunaStatus


def test_code_texts() -> None:
    status_enum_list = list(map(int, SaunaStatus))
    for status_code in status_enum_list:
        assert STATUS_CODE_TEXTS.get(status_code, False)



from typing import Any
from unittest.mock import ANY, patch

import aiohttp
import pytest

from huum.const import SaunaStatus
from huum.exceptions import SafetyException
from huum.huum import Huum
from tests.utils import MockResponse


@pytest.mark.asyncio
async def test_with_session() -> None:
    session = aiohttp.ClientSession()
    Huum("test", "test", session)


@pytest.mark.asyncio
async def test_closing_session() -> None:
    huum = Huum("test", "test")
    await huum.open_session()
    assert huum.session._connector is not None
    await huum.close_session()
    assert huum.session._connector is None


@pytest.mark.asyncio
async def test_no_auth() -> None:
    with pytest.raises(TypeError):
        Huum()  # type: ignore


@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_door_open_on_check(mock_request: Any) -> None:
    response = {
        "statusCode": SaunaStatus.ONLINE_NOT_HEATING,
        "door": False,
        "temperature": 80,
        "maxHeatingTime": 180,
        "saunaName": "test",
    }
    mock_request.return_value = MockResponse(response, 200)

    huum = Huum("test", "test")
    await huum.open_session()
    with pytest.raises(SafetyException):
        await huum._check_door()


@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_bad_temperature_value(mock_request: Any) -> None:
    mock_request.return_value = MockResponse({}, 400)

    huum = Huum("test", "test")
    await huum.open_session()
    with pytest.raises(ValueError):
        await huum.turn_on(temperature=200)


@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_bad_humidity_value(mock_request: Any) -> None:
    mock_request.return_value = MockResponse({}, 400)

    huum = Huum("test", "test")
    await huum.open_session()
    with pytest.raises(ValueError):
        await huum.turn_on(temperature=40, humidity=11)


@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_set_temperature_turn_on(mock_request: Any) -> None:
    response = {
        "statusCode": SaunaStatus.ONLINE_NOT_HEATING,
        "door": True,
        "temperature": 80,
        "maxHeatingTime": 180,
        "saunaName": "test",
    }
    mock_request.return_value = MockResponse(response, 200)

    huum = Huum("test", "test")
    await huum.open_session()
    result_turn_on = await huum.turn_on(temperature=80)
    result_set_temperature = await huum.set_temperature(temperature=80)

    assert result_turn_on == result_set_temperature


@pytest.mark.asyncio
@patch("aiohttp.ClientSession._request")
async def test_humidity_turn_on(mock_request: Any) -> None:
    response = {
        "statusCode": SaunaStatus.ONLINE_NOT_HEATING,
        "door": True,
        "temperature": 80,
        "maxHeatingTime": 180,
        "saunaName": "test",
        "humidity": 5,
    }
    mock_request.return_value = MockResponse(response, 200)

    huum = Huum("test", "test")
    await huum.open_session()
    result_turn_on = await huum.turn_on(temperature=80, humidity=5)

    mock_request.assert_called_with(
        "POST",
        "https://sauna.huum.eu/action/home/start",
        data=None,
        auth=ANY,
        json={"targetTemperature": 80, "humidity": 5},
    )
    assert result_turn_on.humidity == 5



from typing import Any, Optional
from urllib.parse import urljoin

import aiohttp
from aiohttp import ClientResponse

from huum.exceptions import (
    BadRequest,
    Forbidden,
    NotAuthenticated,
    RequestError,
    SafetyException,
)
from huum.schemas import HuumStatusResponse

API_BASE = "https://sauna.huum.eu/action/"
API_HOME_BASE = f"{API_BASE}/home/"


class Huum:
    """
    Usage:
        # Usage with env vars
        huum = Huum()

        # Setting auth variables explicitly
        huum = Huum(username="foo", password="bar")

        # If you don't have an existing aiohttp session
        # then run `open_session()` after initializing
        huum.open_session()

        # Turn on the sauna
        huum.turn_on(80)
    """

    min_temp = 40
    max_temp = 110

    session: aiohttp.ClientSession

    def __init__(
        self,
        username: str,
        password: str,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        if session:
            self.session = session

        self.auth = aiohttp.BasicAuth(username, password)

    async def _check_door(self) -> None:
        """
        Check if the door is closed, if not, raise an exception
        """
        status = await self.status()
        if not status.door_closed:
            raise SafetyException("Can not start sauna when door is open")

    async def _make_call(self, method: str, url: str, json: Any | None = None) -> ClientResponse:
        call_args = {
            "url": url,
            "auth": self.auth,
        }
        if json:
            call_args["json"] = json

        call_request = getattr(self.session, method.lower())

        response: ClientResponse = await call_request(**call_args)

        try:
            response.raise_for_status()
        except Exception as err:
            match response.status:
                case 400:
                    raise BadRequest("Bad request") from err
                case 401:
                    raise NotAuthenticated("Not authenticated") from err
                case 403:
                    raise Forbidden("Forbidden") from err
            raise RequestError() from err

        return response

    async def turn_on(
        self, temperature: int, safety_override: bool = False, humidity: Optional[int] = None
    ) -> HuumStatusResponse:
        """
        Turns on the sauna at a given temperature and optionally humidity

        Args:
            temperature: Target temperature to set the sauna to
            safety_override: If False, check if door is close before turning on the sauna
            humidity: Target humidity duty cycle (optional)

        Returns:
            A `HuumStatusResponse` from the Huum API
        """
        if temperature not in range(self.min_temp, self.max_temp + 1):
            raise ValueError(
                f"Temperature '{temperature}' must be between {self.min_temp}-{self.max_temp}"
            )
        data = {"targetTemperature": temperature}

        if humidity is not None:
            if humidity not in range(0, 11):
                raise ValueError(f"Humidity '{humidity}' must be between 0-10")
            data["humidity"] = humidity

        if not safety_override:
            await self._check_door()

        url = urljoin(API_HOME_BASE, "start")

        response = await self._make_call("post", url, json=data)
        json_data = await response.json()

        return HuumStatusResponse.from_dict(json_data)

    async def turn_off(self) -> HuumStatusResponse:
        """
        Turns off the sauna

        Returns:
            A `HuumStatusResponse` from the Huum API

        """
        url = urljoin(API_HOME_BASE, "stop")

        response = await self._make_call("post", url)
        json_data = await response.json()

        return HuumStatusResponse.from_dict(json_data)

    async def set_temperature(
        self, temperature: int, safety_override: bool = False
    ) -> HuumStatusResponse:
        """
        Alias for turn_on as Huum does not expose an explicit "set_temperature" endpoint

        Implementation choice: Yes, aliasing can be done by simply assigning
        set_temperature = turn_on, however this will not create documentation,
        makes the code harder to read and is generally seen as non-pythonic.

        Args:
            temperature: Target temperature to set the sauna to
            safety_override: If False, check if door is close before turning on the sauna

        Returns:
            A `HuumStatusResponse` from the Huum API
        """
        return await self.turn_on(temperature, safety_override)

    async def status(self) -> HuumStatusResponse:
        """
        Get the status of the Sauna

        Returns:
            A `HuumStatusResponse` from the Huum API
        """
        url = urljoin(API_HOME_BASE, "status")

        response = await self._make_call("get", url)
        json_data = await response.json()

        return HuumStatusResponse.from_dict(json_data)

    async def toggle_light(self) -> HuumStatusResponse:
        """
        Toggles the light/fan on a Sauna

        Returns:
            A `HuumStatusResponse` from the Huum API
        """
        url = urljoin(API_HOME_BASE, "light")

        response = await self._make_call("get", url)
        json_data = await response.json()

        return HuumStatusResponse.from_dict(json_data)

    async def open_session(self) -> None:
        self.session = aiohttp.ClientSession()

    async def close_session(self) -> None:
        await self.session.close()
